/** SoundBit.ino
 **  schip's FHT app used for Sound Control LittleBit by Joey Wilson:
 **  based on openmusiclabs.com 9.5.12 fht_adc.pde
 **   example sketch for testing the fht library.
 **  
 **  Samples ADC0 using schip's ADC ISR code in schip_analog.c
 **  with the SCHIPBUFFER option selected and NUM_ADCCHANNEL == 4
 **  this converts ADC0-3 (with no averaging in ADC code) and
 **  buffers ADC0 for FHT input here.
 **
 **   $Id: Exp $
 **/

// schip's ADC interrupt code
#include "schip_analog.h"
// standard stuff included in above...
// #include <Arduino.h>

/* Set the options for FHT execution -- set lin/log to 1 to use, else 0 */
#define LOG_OUT 0  // 1 == use the log output function, fht_mag_log()
#define LIN_OUT8 1 // 1 == use the lin8 output function, fht_mag_lin8()
#define FHT_N 128 // set to 128 point fht
#include <FHT.h> // include the library
/**/

// Arduino pin assignments
#define LED 13	// main led to flash, usually for signal overload
				//  but may also be set from ADC interrupt for timeing debug
#define LED2 2	// pin 2, another thing to flash, for timing main loop

// PWM outputs
#define LBOUT1	 9	// arduino pin 9
#define LBOUT2	10	// arduino pin 10
#define LBOUT3	11	// arduino pin 11

// other I/O things and stuff
#define OPTIN1	12	// arduino pin 12 -- option select jumper

// output indicators for conversion functions
#define VLOW	1
#define VMID	2
#define VHIGH	3

// The setup routine runs once when you press reset:
// We are presuming that we have all the SCHIP "improvements"
//  to wiring_analog.c for interrupt driven ADC buffering operation
void setup()
{                
	// initialize the led timing tracking pins as an output.
	pinMode(LED, OUTPUT);  
	pinMode(LED2, OUTPUT);  
	// and opition-in pin as input w/ pullup
	pinMode(OPTIN1, INPUT_PULLUP);  

	// initialize the output PWM pins
	//  note: pinmode() is claimed to be redundant when using analogWrite()
	pinMode(LBOUT1, OUTPUT);  
	analogWrite( LBOUT1, 0 );
	pinMode(LBOUT2, OUTPUT);  
	analogWrite( LBOUT2, 0 );
	pinMode(LBOUT3, OUTPUT);  
	analogWrite( LBOUT3, 0 );

	Serial.begin(115200); // use the serial port at high speed for debug

   	// required to start the ADC interrupt cycle
	// which will eventually repeatedly set SchipBUF0
	analogReadMS( 0 );
}

// two kinds of output conversions, see below
extern byte levelOut( char type, byte *outbuf );
extern byte motorOut( char type, byte *outbuf );

/** 64 samples of new ADC0 input data is buffered by the ADC interrupt routine,
 **  when the buffer is full it is passed in here when ISR sets SchipBUF0
 **  to the address of the buffer of 64 (SBUFSIZE) 10bit (word) values.
 ** The ADC interrupt routine then starts filling a second buffer.
 ** This routine copies the new data into the upper half of the FHT input array
 **  and runs the FHT. When done this current data is copied into the lower
 **  half of the input array, where new data is concatenated on the next pass.
 ** This has to happen before the new ADC buffering completes...
 ** The results of the FHT are then used to set the PWM duty cycle on the
 **  three output pins.
 **
 **  For the PRO-Mini (and seemingly the regular Arduino) with
 **   a 16Mhz Atmega 328, and the ADC prescale set at maximum 128 clocks:
 **		A single ADC conversion takes about .110ms
 **		A four channel ADC cycle takes about .45ms (2222Hz sample rate)
 **		64 conversion cycles takes about 28.8ms
 ** 	  -- which is the buffer update period when (SchipBUF0 != NULL) --
 **		and the 128 point FHT takes about 2.5ms so we are golden
 **
 **/ 
void loop()
{
	static byte x = 0;		// counter for debug printing
	static byte dbg = 0;	// command for debug printing
	byte h, m, l;			// actual PWM setting values
	byte *outbuf;			// output buffer we use

	// get some debug commands,
	// a single character with no <CR> or anything...
	if( Serial.available() > 0 )
	{
		// get incoming byte
		byte inByte = Serial.read();
		//Serial.println(inByte);

		x = 0;	// restart counter so we delay before debug printing

		// 0 shuts off debug printing
		// 1 will print input array
		// 2 will print FHT output
		// 3 will print PWM settings only
		if(      inByte == '1' )	dbg = 1;
		else if( inByte == '2' )	dbg = 2;
		else if( inByte == '3' )	dbg = 3;
		else						dbg = 0;
    }

	// wait for data to be available
	//  SchipBUF0 is set to a new buffer address by interrupt routine
	if( SchipBUF0 != NULL )
    {
		// note: LED may interfere with ADC ISR tracing, cavet emptor
  	    digitalWrite(LED, 0);   // wiggle LED pin OFF for signal level trace
  	    digitalWrite(LED2, 1);   // wiggle LED2 pin ON to trace execution
		++x;	// debug print counter

		/** Get new sample data into upper half of FHT input buffer **/
		// todo: to save uS, could keep new ints in current SchipBUF,
		//  and do a memcpy() when resetting input conversion buffer below...
		for( int i = 0 ; i < SBUFSIZE ; ++i ) // copy samples
		{
			int k = SchipBUF0[i]; // get ADC data
			k -= 0x0200; // make a signed int around nominal zero point
			// promote to 16b signed int and save in input bins
			fht_input[i+SBUFSIZE] = k << 6;
		}
		/** end copy new data **/

		/** debug print the input data once in a while (~7.3 sec) **/
		if( (dbg == 1) && (x == 64) )
		{
			Serial.println( "Input data:" );
			for( int i = 0 ; i < FHT_N ; ++i )
				Serial.println( fht_input[i] );
			Serial.println( "" );
		}
		/* end debug printf */

		/** now do the FHT deed **/
		fht_window();	// window the data for better frequency response
		fht_reorder();	// reorder the data before doing the fht
		fht_run();		// process the data in the fht
		// calculate real frequency-level values
#if LOG_OUT == 1  // 1 == use the log output function, fht_mag_log()
    fht_mag_log(); // take the log output of the fht
	outbuf = fht_log_out;
#endif
#if LIN_OUT8 == 1 // 1 == use the lin8 output function, fht_mag_lin8()
	fht_mag_lin8(); // take the linear output of the fht
	outbuf = fht_lin_out8;
#endif

		/** Put current sample data into lower half of input buffer,
		 **  new data will be concatenated to it on next pass **/
		// todo: could keep new ints in current SchipBUF then memcpy()
		//       -- but don't really need to speed things up -- so what.
		for( int i = 0 ; i < 64 ; ++i ) // copy samples
		{
			// form into 16bit signed int and
			// put in lower half of input bins
			int k = SchipBUF0[i];
			k -= 0x0200;
			fht_input[i] = k << 6;
		}
		/** end copy current data for next time **/

		// clear the semaphore --
		//  hopefully this all happens before the new sample buffer is ready
		SchipBUF0 = NULL;

		// now convert to output values
		if( digitalRead( OPTIN1 ) == 0 )
		{
		//Serial.println("MOTOR");
			// pin12 jumper is ON -- do MOTOR type levels
			l = motorOut( VLOW, outbuf );
			m = motorOut( VMID, outbuf );
			h = motorOut( VHIGH, outbuf );
		}
		else
		{
		//Serial.println("LEVEL");
			// pin12 jumper is OFF -- do direct value type levels
			l = levelOut( VLOW, outbuf );
			m = levelOut( VMID, outbuf );
			h = levelOut( VHIGH, outbuf );
		}

		analogWrite( LBOUT1, l ); // low to output 1
		analogWrite( LBOUT2, m ); // middle to output 2
		analogWrite( LBOUT3, h ); // high to output 3

		digitalWrite(LED2, 0);   // wiggle LED2 OFF for execution timing trace

		/** debug print the FHT or PWM data once in a while (~7.3 sec) **/
		if( (dbg == 2) && (x == 64) )
		{
			Serial.println( "FHT output:" );
			for( int i = 0 ; i < FHT_N/2 ; ++i )
				Serial.println( outbuf[i] );
			Serial.print( "PWM: " );
			Serial.print( l );
			Serial.print( ", " );
			Serial.print( m );
			Serial.print( ", " );
			Serial.println( h );
			Serial.println( "" );
		}
		/* end debug print */

		/** debug print the PWM data once in a while (~2 sec) **/
		if( (dbg == 3) && (x == 64) )
		{
			x = 0;	// reset counter so this happens about every 2 sec
			Serial.print( "PWM: " );
			Serial.print( l );
			Serial.print( ", " );
			Serial.print( m );
			Serial.print( ", " );
			Serial.println( h );
		}
		/* end debug print */

	} // end'o'doingFHT

} // end'o'loop

/** two kinds of output conversion...
 **
 ** @param freq   -- frequency of interest HIGH, MIDDLE, LOW
 ** @param outbuf -- FHT output buffer -- presumed to be bytes in LIN8 format
 **
 ** @return -- level to set on specified output, 0-255
 **            also sets LED (p13) ON if the input signal is at max value
 **
 **  levelOut() -- uses strength of center bin for each frequency
 **					to set the output level directly
 **  motorOut() -- uses side bins for each freq to set output level +/-,
 **					so the no-signal value is mid-range (128)
 **					a signal on the higher bin raises the value proportionately
 **					and a signal on the lower bin reduces the value inversely.
 **/

/** at .45ms sample rate bins 13, 34, 57 are notes A3, D5, B5
 ** For LOG output, values range (noise floor to peak):
 **  from about 20-190 with 'clean' full scale simulated data
 **  with 'real' input the noise floor is more like 60-80
 ** For LIN8 output, the range seems to be 0-60 with real input
 **  but need to keep individual notes around level 25 so 
 **  the combined signal stays below the clipping threshold.
 ** We are using LIN8 output ...
**/

/** Useful note frequencies:
oct   C     C#    D     Eb    E     F     F#    G     G#    A     Bb    B
 0	16.35 17.32 18.35 19.45 20.60 21.83 23.12 24.50 25.96 27.50 29.14 30.87
 1	32.70 34.65 36.71 38.89 41.20 43.65 46.25 49.00 51.91 55.00 58.27 61.74
 2	65.41 69.30 73.42 77.78 82.41 87.31 92.50 98.00 103.8 110.0 116.5 123.5
 3	130.8 138.6 146.8 155.6 164.8 174.6 185.0 196.0 207.7 220.0 233.1 246.9
 4	261.6 277.2 293.7 311.1 329.6 349.2 370.0 392.0 415.3 440.0 466.2 493.9
 5	523.3 554.4 587.3 622.3 659.3 698.5 740.0 784.0 830.6 880.0 932.3 987.8
 6	1047  1109  1175  1245  1319  1397  1480  1568  1661  1760  1865  1976
 7	2093  2217  2349  2489  2637  2794  2960  3136  3322  3520  3729  3951
 8	4186  4435  4699  4978  5274  5588  5920  6272  6645  7040  7459  7902
**/

// bin locations, numbered from 0-63
#define LDB 13	// 220 hz -- A3
#define MDB 34	// 587 hz -- D5
#define HDB 57	// 988 hz -- B5

// measured high/low bin numbers
#define LDBl 11	// 196hz -- G3
#define LDBh 14 // 246hz -- B3
#define MDBl 30	// 523hz -- C5
#define MDBh 38 // 659hz -- E5
#define HDBl 53	// 932hz -- Bb5
#define HDBh 60 // 1047hz -- C6

// signal stuff
#define NFLOOR 3	// noise floor
#define SMULT  13	// multiplier to make it to 255 max
#define MIDVAL 128	// middle output value for motorOut()

byte levelOut( char type, byte *outbuf )
{
	int lev;

	// get the FHT level of interest
	if(      type == VLOW )	{ lev = outbuf[LDB]; }
	else if( type == VMID ) { lev = outbuf[MDB]; }
	else                    { lev = outbuf[HDB]; }

	// subtract noise then adjust and clamp value 
	lev -= NFLOOR;
	if( lev <= 0 )	lev = 0;
	else			lev *= SMULT;

	if( lev >= 255 )
	{
		lev = 255;
		digitalWrite(LED, 1); // wiggle LED pin ON for signal level trace
	}

	return (byte) lev;
}

byte motorOut( char type, byte *outbuf )
{
	int lev = MIDVAL*2;	// use *2 so we can use same SMULT as levelOut()
	byte levl, levh;

	// get the FHT levels of interest
	if(      type == VLOW )	{ levl = outbuf[LDB-2]; levh = outbuf[LDB+1]; }
	else if( type == VMID ) { levl = outbuf[MDB-4]; levh = outbuf[MDB+4]; }
	else 					{ levl = outbuf[HDB-4]; levh = outbuf[HDB+3]; }

	// which one is bigger?
	if( levl > levh )
	{
		// low bin, invert sense to subtract from MIDVAL
		// subtract noise then adjust and clamp value 
		lev = levl - NFLOOR;
		if( lev <= 0 )	lev = 0;
		else			lev *= SMULT;

		if( lev >= 255 )
		{
			lev = 255;
			digitalWrite(LED, 1); // wiggle LED pin ON for signal level trace
		}

		lev = (MIDVAL-1) - (lev >> 1);
	}
	else
	{
		// high bin, add to MIDVAL
		// subtract noise then adjust and clamp value 
		lev = levh - NFLOOR;
		if( lev <= 0 )	lev = 0;
		else			lev *= SMULT;

		if( lev >= 255 )
		{
			lev = 255;
			digitalWrite(LED, 1); // wiggle LED pin ON for signal level trace
		}

		lev = MIDVAL + (lev >> 1);
	}

	return (byte) lev;
}
